#![no_std]
#![no_main]
use panic_halt as _;

#[rtic::app(device = waveshare_rp2040_zero::hal::pac, peripherals = true, dispatchers = [I2C0_IRQ])]
mod app {
    use core::convert::TryInto;
    use core::iter::once;
    use cortex_m::delay::Delay;
    use fugit::MicrosDurationU32;
    use fugit::RateExtU32;
    use rp2040_hal::gpio::bank0::{Gpio16, Gpio2, Gpio3, Gpio4, Gpio5};
    use rp2040_hal::gpio::{
        FunctionPio0, FunctionSio, FunctionSpi, Pin, PullDown, PullNone, PullUp, SioOutput,
    };
    use rp2040_hal::pio::SM0;
    use rp2040_hal::timer::CountDown;
    use rp2040_hal::{Clock, Spi};
    use smart_leds::{brightness, SmartLedsWrite};
    use usbd_serial::SerialPort;
    use waveshare_rp2040_zero::hal::prelude::*;
    use waveshare_rp2040_zero::hal::spi::Enabled;
    use waveshare_rp2040_zero::{
        hal::{
            self, clocks::init_clocks_and_plls, timer::monotonic::Monotonic, timer::Alarm,
            watchdog::Watchdog, Sio,
        },
        pac::{PIO0, SPI0},
        XOSC_CRYSTAL_FREQ,
    };

    use usb_device::{class_prelude::*, prelude::*};
    use waveshare_rp2040_zero::hal::pio::PIOExt;
    use ws2812_pio::Ws2812;

    use super::*;
    use core::fmt::Write;
    use embedded_hal::digital::OutputPin;
    use fugit::ExtU32;
    use rp2040_hal::timer::Alarm0;
    use w5500_dhcp::{
        hl::Hostname,
        ll::{
            eh0::vdm_infallible_gpio::W5500,
            eh0::MODE as W5500_MODE,
            net::{Eui48Addr, Ipv4Addr, SocketAddrV4},
            LinkStatus, OperationMode, PhyCfg, Registers, Sn,
        },
        Client as DhcpClient,
    };
    use w5500_mqtt::{
        Client as MqttClient, ClientId, Event as MqttEvent, SRC_PORT as MQTT_SRC_PORT,
    };
    const SYSCLK_HZ: u32 = 8_000_000;

    const SCAN_TIME_US: MicrosDurationU32 = MicrosDurationU32::millis(400);

    const DHCP_SN: Sn = Sn::Sn0;
    const MQTT_SN: Sn = Sn::Sn1;

    const MQTT_SERVER: SocketAddrV4 = SocketAddrV4::new(Ipv4Addr::new(192, 168, 0, 14), 1883);
    const NAME: &str = "ambient1";
    const HOSTNAME: Hostname<'static> = Hostname::new_unwrapped(NAME);
    const CLIENT_ID: ClientId<'static> = ClientId::new_unwrapped(NAME);

    #[monotonic(binds = TIMER_IRQ_0, default = true)]
    type MyMono = Monotonic<Alarm0>;

    #[inline(always)]
    pub fn delay_ms(ms: u32) {
        const CYCLES_PER_MILLIS: u32 = SYSCLK_HZ / 1000;
        cortex_m::asm::delay(CYCLES_PER_MILLIS.saturating_mul(ms));
    }

    fn monotonic_secs() -> u32 {
        app::monotonics::now()
            .duration_since_epoch()
            .to_secs()
            .try_into()
            .unwrap()
    }
    #[shared]
    struct Shared {
        usb_dev: UsbDevice<'static, hal::usb::UsbBus>,
        serial: SerialPort<'static, hal::usb::UsbBus>,
        ws: Ws2812<PIO0, SM0, CountDown<'static>, hal::gpio::Pin<Gpio16, FunctionPio0, PullDown>>,
        w5500: W5500<
            Spi<
                Enabled,
                SPI0,
                (
                    Pin<Gpio3, FunctionSpi, PullNone>,
                    Pin<Gpio4, FunctionSpi, PullUp>,
                    Pin<Gpio2, FunctionSpi, PullNone>,
                ),
            >,
            Pin<Gpio5, FunctionSio<SioOutput>, PullDown>,
        >,
        dhcp: DhcpClient<'static>,
        mqtt: MqttClient<'static>,
        dhcp_spawn_at: Option<u32>,
        mqtt_spawn_at: Option<u32>,
    }

    #[local]
    struct Local {}

    #[init]
    fn init(c: init::Context) -> (Shared, Local, init::Monotonics) {
        // Soft-reset does not release the hardware spinlocks
        // Release them now to avoid a deadlock after debug or watchdog reset
        unsafe {
            hal::sio::spinlock_reset();
        }
        let mut resets = c.device.RESETS;
        let mut watchdog = Watchdog::new(c.device.WATCHDOG);
        let clocks = init_clocks_and_plls(
            XOSC_CRYSTAL_FREQ,
            c.device.XOSC,
            c.device.CLOCKS,
            c.device.PLL_SYS,
            c.device.PLL_USB,
            &mut resets,
            &mut watchdog,
        )
        .ok()
        .unwrap();

        let sio = Sio::new(c.device.SIO);
        let pins = waveshare_rp2040_zero::Pins::new(
            c.device.IO_BANK0,
            c.device.PADS_BANK0,
            sio.gpio_bank0,
            &mut resets,
        );

        // Set up our SPI pins into the correct mode
        let spi_sclk: hal::gpio::Pin<_, hal::gpio::FunctionSpi, hal::gpio::PullNone> =
            pins.gp2.reconfigure();
        let spi_mosi: hal::gpio::Pin<_, hal::gpio::FunctionSpi, hal::gpio::PullNone> =
            pins.gp3.reconfigure();
        let spi_miso: hal::gpio::Pin<_, hal::gpio::FunctionSpi, hal::gpio::PullUp> =
            pins.gp4.reconfigure();
        let spi_cs = pins.gp5.into_push_pull_output();
        let mut spi_reset = pins.gp7.into_push_pull_output();

        // Create the SPI driver instance for the SPI0 device
        let spi = hal::spi::Spi::<_, _, _, 8>::new(c.device.SPI0, (spi_mosi, spi_miso, spi_sclk));
        // Exchange the uninitialised SPI driver for an initialised one
        let spi = spi.init(
            &mut resets,
            clocks.peripheral_clock.freq(),
            1.MHz(),
            W5500_MODE,
        );
        let mut w5500 = W5500::new(spi, spi_cs);

        w5500_dhcp::ll::eh0::reset(&mut spi_reset, &mut Delay::new(c.core.SYST, SYSCLK_HZ))
            .unwrap();

        //let timer = Timer::new(c.device.TIMER, &mut c.device.RESETS, &clocks);
        let dTIMER = c.device.TIMER;
        let timer =
            cortex_m::singleton!(: hal::Timer = hal::Timer::new(dTIMER, &mut resets, &clocks))
                .unwrap();
        let mut alarm = timer.alarm_0().unwrap();

        let count_down: CountDown<'static> = timer.count_down();

        let (mut pio, sm0, _, _, _) = c.device.PIO0.split(&mut resets);
        let ws: Ws2812<
            PIO0,
            SM0,
            CountDown<'static>,
            hal::gpio::Pin<Gpio16, FunctionPio0, PullDown>,
        > = Ws2812::new(
            // The onboard NeoPixel is attached to GPIO pin #16 on the Waveshare RP2040-Zero.
            pins.neopixel.into_function(),
            &mut pio,
            sm0,
            clocks.peripheral_clock.freq(),
            count_down,
        );

        // Set up the USB driver
        let USBCTRL_REGS = c.device.USBCTRL_REGS;
        let USBCTRL_DPRAM = c.device.USBCTRL_DPRAM;
        let usb_clock = clocks.usb_clock;
        let usb_bus = cortex_m::singleton!(
            : UsbBusAllocator<hal::usb::UsbBus> = UsbBusAllocator::new(waveshare_rp2040_zero::hal::usb::UsbBus::new(
            USBCTRL_REGS,
            USBCTRL_DPRAM,
            usb_clock,
            true,
            &mut resets,
        ))).unwrap();

        // Set up the USB Communications Class Device driver
        let mut serial = SerialPort::new(usb_bus);

        // Create a USB device with a fake VID and PID
        let usb_dev = UsbDeviceBuilder::new(usb_bus, UsbVidPid(0x16c0, 0x27dd))
            .strings(&[StringDescriptors::default()
                .manufacturer("Fake company")
                .product("Serial port")
                .serial_number("TEST")])
            .unwrap()
            .device_class(2) // from: https://www.usb.org/defined-class-codes
            .build();

        let _ = serial.write(b"Hello, World!\r\n");

        let _ = alarm.schedule(SCAN_TIME_US);
        alarm.enable_interrupt();

        let mac: Eui48Addr = Eui48Addr::new(0x00, 0x08, 0xDC, 0x00, 0x00, 0x00);

        timer_irq::spawn().unwrap();

        // continually initialize the W5500 until we link up
        // since we are using power over Ethernet we know that if the device
        // has power it also has an Ethernet cable connected.
        let _phy_cfg: PhyCfg = 'outer: loop {
            // sanity check W5500 communications
            assert_eq!(w5500.version().unwrap(), w5500_dhcp::ll::VERSION);

            // load the MAC address we got from EEPROM
            w5500.set_shar(&mac).unwrap();
            debug_assert_eq!(w5500.shar().unwrap(), mac);

            // wait for the PHY to indicate the Ethernet link is up
            let mut attempts: u32 = 0;
            serial.write(b"Polling for link up\r\n");
            const PHY_CFG: PhyCfg = PhyCfg::DEFAULT.set_opmdc(OperationMode::FullDuplex10bt);
            w5500.set_phycfgr(PHY_CFG).unwrap();

            const LINK_UP_POLL_PERIOD_MILLIS: u32 = 100;
            const LINK_UP_POLL_ATTEMPTS: u32 = 50;
            loop {
                let phy_cfg: PhyCfg = w5500.phycfgr().unwrap();
                if phy_cfg.lnk() == LinkStatus::Up {
                    break 'outer phy_cfg;
                }
                if attempts >= LINK_UP_POLL_ATTEMPTS {
                    let mut data: heapless::String<16> = heapless::String::new();
                    write!(
                        &mut data,
                        "Failed to link up in {} ms\r\n",
                        attempts * LINK_UP_POLL_PERIOD_MILLIS
                    )
                    .unwrap();

                    serial.write(data.as_bytes());
                    break;
                }
                delay_ms(LINK_UP_POLL_PERIOD_MILLIS);
                attempts += 1;
            }

            spi_reset.set_low().unwrap();
            delay_ms(1);
            spi_reset.set_high().unwrap();
            delay_ms(3);
        };
        let mut data: heapless::String<16> = heapless::String::new();
        write!(&mut data, "Done link up\n{}", _phy_cfg).unwrap();

        serial.write(data.as_bytes());

        let mut mqtt: MqttClient = MqttClient::new(MQTT_SN, MQTT_SRC_PORT, MQTT_SERVER);
        mqtt.set_client_id(CLIENT_ID);

        let seed: u64 = u64::from(cortex_m::peripheral::SYST::get_current()) << 32
            | u64::from(cortex_m::peripheral::SYST::get_current());

        let dhcp = DhcpClient::new(DHCP_SN, seed, mac, HOSTNAME);
        dhcp.setup_socket(&mut w5500).unwrap();

        // start the DHCP client
        //dhcp_sn::spawn().unwrap();

        // start the timeout tracker
        //timeout_tracker::spawn().unwrap();

        (
            Shared {
                usb_dev,
                serial,
                ws,
                w5500,
                dhcp,
                mqtt,
                dhcp_spawn_at: None,
                mqtt_spawn_at: None,
            },
            Local {},
            init::Monotonics(Monotonic::new(*timer, alarm)),
        )
    }

    #[task(shared = [dhcp_spawn_at, mqtt_spawn_at])]
    fn timeout_tracker(mut cx: timeout_tracker::Context) {
        timeout_tracker::spawn_after(1_u32.secs().into()).unwrap();

        let now: u32 = monotonic_secs();

        cx.shared.dhcp_spawn_at.lock(|dhcp_spawn_at| {
            if let Some(then) = dhcp_spawn_at {
                if now >= *then {
                    if dhcp_sn::spawn().is_err() {
                        log::error!("DHCP task is already spawned")
                    }
                    *dhcp_spawn_at = None;
                }
            }
        });

        cx.shared.mqtt_spawn_at.lock(|mqtt_spawn_at| {
            if let Some(then) = mqtt_spawn_at {
                if now >= *then {
                    if mqtt_sn::spawn().is_err() {
                        log::error!("MQTT task is already spawned")
                    }
                    *mqtt_spawn_at = None;
                }
            }
        });
    }

    #[task(shared = [w5500, mqtt, mqtt_spawn_at], local = [])]
    fn mqtt_sn(cx: mqtt_sn::Context) {
        log::info!("[TASK] mqtt_sn");

        (cx.shared.w5500, cx.shared.mqtt, cx.shared.mqtt_spawn_at).lock(
            |w5500, mqtt, mqtt_spawn_at| {
                loop {
                    let now: u32 = monotonic_secs();
                    match mqtt.process(w5500, now) {
                        Ok(MqttEvent::CallAfter(secs)) => {
                            *mqtt_spawn_at = Some(now + secs);
                            break;
                        }
                        Ok(MqttEvent::ConnAck) => {
                            log::info!("[MQTT] ConnAck");
                            // can subscribe to topics here
                            // not needed for this
                        }
                        Ok(MqttEvent::Publish(reader)) => {
                            log::warn!("should not get Publish never subscribed");
                            reader.done().unwrap();
                        }
                        Ok(MqttEvent::SubAck(_) | MqttEvent::UnSubAck(_)) => {
                            log::warn!("should not get (Un)SubAck, never (un)subscribed");
                        }
                        Ok(MqttEvent::None) => {
                            {
                                let mut data: heapless::String<16> = heapless::String::new();
                                write!(&mut data, "{:.1}", 42).unwrap();
                                mqtt.publish(w5500, "test", data.as_bytes()).unwrap();
                            }
                            *mqtt_spawn_at = Some(now + 5);
                            break;
                        }
                        Err(e) => {
                            log::error!("[MQTT] {e:?}");
                            *mqtt_spawn_at = Some(now + 10);
                            break;
                        }
                    }
                }
            },
        );
    }

    #[task(shared = [w5500, dhcp, dhcp_spawn_at])]
    fn dhcp_sn(cx: dhcp_sn::Context) {
        log::info!("[TASK] dhcp_sn");

        (cx.shared.w5500, cx.shared.dhcp, cx.shared.dhcp_spawn_at).lock(
            |w5500, dhcp, dhcp_spawn_at| {
                let leased_before: bool = dhcp.has_lease();
                let now: u32 = monotonic_secs();
                let spawn_after_secs: u32 = dhcp.process(w5500, now).unwrap();

                let spawn_at: u32 = now + spawn_after_secs;
                *dhcp_spawn_at = Some(spawn_at);
                log::info!("[DHCP] spawning after {spawn_after_secs} seconds, at {spawn_at}");

                // spawn MQTT task if bound
                if dhcp.has_lease() && !leased_before && mqtt_sn::spawn().is_err() {
                    log::error!("MQTT task is already spawned")
                }
            },
        )
    }

    #[task(
        priority = 1,
        shared = [serial, ws],
        local = [tog: bool = true],
    )]
    fn timer_irq(mut c: timer_irq::Context) {
        if *c.local.tog {
            c.shared
                .ws
                .lock(|ws| ws.write(brightness(once((255, 0, 0).into()), 32)).unwrap());
            c.shared.serial.lock(|serial| _ = serial.write(b"High\r\n"));
        } else {
            c.shared
                .ws
                .lock(|ws| ws.write(brightness(once((0, 0, 0).into()), 32)).unwrap());
            c.shared.serial.lock(|serial| _ = serial.write(b"LOW\r\n"));
        }
        *c.local.tog = !*c.local.tog;

        timer_irq::spawn_after(500.millis().into()).unwrap();
    }

    #[task(
        binds = USBCTRL_IRQ,
        priority = 1,
        shared = [usb_dev, serial],
    )]
    fn usbctrl_irq(c: usbctrl_irq::Context) {
        // Grab the global objects. This is OK as we only access them under interrupt.
        let usb_dev = c.shared.usb_dev;
        let serial = c.shared.serial;

        (usb_dev, serial).lock(|usb_dev, serial| {
            // Poll the USB driver with all of our supported USB Classes

            if usb_dev.poll(&mut [serial]) {
                let mut buf = [0u8; 64];
                match serial.read(&mut buf) {
                    Err(_e) => {
                        // Do nothing
                    }
                    Ok(0) => {
                        // Do nothing
                    }
                    Ok(count) => {
                        // Convert to upper case
                        buf.iter_mut().take(count).for_each(|b| {
                            b.make_ascii_uppercase();
                        });

                        // Send back to the host
                        let mut wr_ptr = &buf[..count];
                        while !wr_ptr.is_empty() {
                            let _ = serial.write(wr_ptr).map(|len| {
                                wr_ptr = &wr_ptr[len..];
                            });
                        }
                    }
                }
            }
        });
    }

    /// This is the W5500 interrupt.
    ///
    /// The only interrupts we should get are for the DHCP & MQTT sockets.
    #[task(binds = SPI0_IRQ, local = [], shared = [w5500])]
    #[allow(clippy::collapsible_if)]
    fn exti0(mut cx: exti0::Context) {
        log::info!("[TASK] exti0");
    }
}
